
#define WIN32_LEAN_AND_MEAN

#include "ftc.h"

#ifdef TEC_WINDOWS
#include <winsock2.h>
#include <windows.h>
#else
#include <inttypes.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#endif

#include <stdexcept>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <algorithm>

#include <vector>

namespace ftc_detail {

#ifdef TEC_WINDOWS
typedef SOCKET socket_type;
#define TEC_INVALID_SOCKET INVALID_SOCKET
#define TEC_SOCKET_ERROR SOCKET_ERROR
#else
typedef int socket_type;
#define TEC_INVALID_SOCKET (-1)
#define TEC_SOCKET_ERROR (-1)
#endif

namespace {

#ifdef TEC_WINDOWS
  const std::size_t win32_minimum_buffer_size(16);
#endif

void send(socket_type socket, const char* buffer, std::size_t size);
void recv(socket_type socket, char* buffer, std::size_t size);

struct errno_buffer_ref
{
  errno_buffer_ref(const char* buffer)
    : buffer(buffer) {}
  const char* buffer;
};

// errno_buffer has move semantics
struct errno_buffer
{
  errno_buffer(errno_buffer_ref r)
    : buffer(r.buffer) {}
  explicit errno_buffer(const char* buffer) : buffer(buffer) {}
  errno_buffer(errno_buffer& o)
    : buffer(o.release())
  {}
  errno_buffer& operator=(errno_buffer& o)
  {
    buffer = o.release();
    return *this;
  }
#ifdef TEC_WINDOWS
  ~errno_buffer()
  {
    if(buffer)
      ::LocalFree(static_cast<HLOCAL>
                  (const_cast<void*>
                   (static_cast<const void*>(buffer))));
  }
#endif
  operator errno_buffer_ref ()
  {
    return errno_buffer_ref(release());
  }
  const char* message() const { return buffer; }
  const char* release()
  {
    const char* b = buffer;
    buffer = 0;
    return b;
  }
private:
  const char* buffer;
};

errno_buffer errno_message()
{
#ifdef linux
  return errno_buffer(sys_errlist[errno]);
#elif defined(TEC_WINDOWS)
  const char* buffer = 0;
  DWORD r = ::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER
                            | FORMAT_MESSAGE_FROM_SYSTEM
                            | FORMAT_MESSAGE_IGNORE_INSERTS
                            , 0, WSAGetLastError()
                            , 0, static_cast<LPTSTR>(static_cast<void*>(&buffer))
                            , 0, 0);
  if(buffer)
    return errno_buffer(buffer);
  else
  {
    std::abort();
    throw -1;
  }
#else
  return errno_buffer("A system error");
#endif
}

void close(socket_type s)
{
#ifdef TEC_WINDOWS
  ::closesocket(s);
#else
  ::close(s);
#endif  
}

struct socket
{
  socket(int s)
    : sock(s) {}

  ~socket()
  {
    if(sock != TEC_INVALID_SOCKET)
      close(sock);
  }

  int release() 
  {
    int s = sock;
    sock = TEC_INVALID_SOCKET; 
    return s; 
  }
  int get() const { return sock; }
  
  int sock;
};

void throw_exception(int status, const char* message)
{
  switch(status)
  {
  case -1:
    throw FailureException(message);
  case -2:
    throw InvalidKeyException(message);
  case -3:
    throw FileNotFoundException(message);
  case -4:
    throw NoPermissionException(message);
  case -5:
    throw FileLockedException(message);
  case -6:
    throw MaxClientsReachedException(message);
  case -7:
    throw FileNotOpenException(message);
  default:
    throw FtcException(static_cast<FtcErrorCode>(static_cast<unsigned char>(static_cast<char>(status))), message);
  }
}

void throw_exception(const char* message)
{
  throw FailureException(message);
}

int recv_status(socket_type socket)
{
#ifdef TEC_WINDOWS
  char status = 0;
  WSABUF buffers = {1, &status};
  DWORD count = 0, flags = 0;
  int r = WSARecv(socket, &buffers, 1, &count, &flags
                  , 0, 0);
  if(r != 0)
  {
    errno_buffer b = errno_message();
    throw_exception(b.message());
  }
  else if(count == 0)
    throw_exception("EOF");
  return status;
#else
  char status = 0;
  ssize_t r = ::recv(socket, &status, 1, 0);
  if(r == -1)
  {
    errno_buffer b = errno_message();
    throw_exception(b.message());
  }
  else if(r == 0)
    throw_exception("EOF");
  return status;
#endif
}

#if defined(LITTLE_ENDIAN) || defined(__LITTLE_ENDIAN) || defined(__LITTLE_ENDIAN__) || defined(TEC_WINDOWS)
void endian_swap(char* buffer, std::size_t size)
{
  std::vector<char> reverse_endian(size);
  std::copy(buffer, buffer + size, reverse_endian.rbegin());
  std::memcpy(buffer, &reverse_endian[0], size);
}
#else
void endian_swap(char*, std::size_t) {}
#endif

unsigned long long return_long_long_command(socket_type socket, char command)
{
  // This is a compile-time assertion that unsigned long long is 8 bytes
  char v[sizeof(unsigned long long) == 8?1:0];
  (void)v;

  send(socket, &command, 1);

  char recv_buffer[8];
  recv(socket, recv_buffer, sizeof(recv_buffer));

  endian_swap(recv_buffer, 8);

  unsigned long long return_value;
  std::memcpy(&return_value, recv_buffer
              , (std::min)(sizeof(return_value),sizeof(recv_buffer)));
  return return_value;
}

void send(socket_type socket, const char* buffer, std::size_t size)
{
  std::size_t total = 0;
  while(total != size)
  {
#ifdef TEC_WINDOWS
    WSABUF buffers = {size, const_cast<char*>(buffer)};
    DWORD count = 0;
    int r = WSASend(socket, &buffers, 1, &count, 0, 0, 0);
    if(r != 0)
    {
      errno_buffer b = errno_message();
      throw_exception(b.message());
    }
    else if(count == 0)
      throw_exception("EOF");
    total += count;
#else
    ssize_t r = 0;
    do
    {
      r = ::send(socket, &buffer[total], size-total, 0);
    } while(r == -1 && errno == EINTR);
    if(r == -1)
    {
      errno_buffer b = errno_message();
      throw_exception(b.message());
    }
    else if(r == 0)
      throw_exception("EOF");
    total += r;
#endif
  }
}

void recv(socket_type socket, char* buffer, std::size_t size)
{
  std::size_t total = 0;
  while(total != size)
  {
#ifdef TEC_WINDOWS
    WSABUF buffers = {size, buffer};
    DWORD count = 0;
    DWORD flags = 0;
    int r = ::WSARecv(socket, &buffers, 1, &count, &flags, 0, 0);
    if(r != 0)
    {
      errno_buffer b = errno_message();
      throw_exception(b.message());
    }
    else if(count == 0)
      throw_exception("EOF");
    total += count;
#else
    ssize_t r = 0;
    do
    {
      r = ::recv(socket, &buffer[total], size-total, 0);
    } while(r == -1 && errno == EINTR);
    if(r == -1)
    {
      errno_buffer b = errno_message();
      ftc_detail::throw_exception(b.message());
    }
    else if(r == 0)
      ftc_detail::throw_exception("EOF");
    total += r;
#endif
  }
}


}

}

ftc::ftc(const char* id, const char id_size, bool writable, const char* host
        , unsigned short port, const char* accessKey, const char key_size)
  : hostname(host)
  , tcp_port(port)
  , identifier (id)
  , identifier_size (id_size)
  , sock(TEC_INVALID_SOCKET)
  , writable(writable)
  , is_open(false)
  , access_key(accessKey)
  , access_key_size(key_size)
{
}

void ftc::open(bool readonly)
{
  if(!writable && !readonly)
    throw NoPermissionException("Permission set in constructor");

  writable = !readonly;

#ifdef TEC_WINDOWS
  WORD wVersionRequested = MAKEWORD(2,2);
  WSADATA wsaData;
  int err = WSAStartup(wVersionRequested, &wsaData);
  if(err != 0)
  {
    ftc_detail::errno_buffer b = ftc_detail::errno_message();
    ftc_detail::throw_exception(b.message());
  }
#endif  
  ftc_detail::socket s(::socket(AF_INET, SOCK_STREAM, 0));
  if(s.get() == TEC_INVALID_SOCKET)
  {
    ftc_detail::errno_buffer b = ftc_detail::errno_message();
    ftc_detail::throw_exception(b.message());
  }

  hostent* h = gethostbyname(hostname);

  if(!h)
  {
    ftc_detail::errno_buffer b = ftc_detail::errno_message();
    ftc_detail::throw_exception(b.message());
  }

  ::sockaddr_in addr = 
      {AF_INET, htons(tcp_port)}; // htons can't be qualified, because it can be a macro

  std::memcpy(&addr.sin_addr.s_addr, h->h_addr_list[0], sizeof(addr.sin_addr.s_addr));

  if(::connect(s.get(), static_cast<struct sockaddr*>(static_cast<void*>(&addr))
               , sizeof addr) == TEC_SOCKET_ERROR)
  {
    ftc_detail::errno_buffer b = ftc_detail::errno_message();
    ftc_detail::throw_exception(b.message());
  }

  sock = s.release();

  // authenticate
  {
    char size = access_key_size;
    ftc_detail::send(sock, &size, 1);
    ftc_detail::send(sock, access_key, access_key_size);
    int status = ftc_detail::recv_status(sock);
    if(status != 0)
      ftc_detail::throw_exception(status, "Authentication");
  }

  // open
  char buffer[2];
  buffer[0] = writable;
  buffer[1] = identifier_size;
  ftc_detail::send(sock, buffer, sizeof(buffer));
  ftc_detail::send(sock, identifier, buffer[1]);
  int status = ftc_detail::recv_status(sock);
  if(status != 0)
    ftc_detail::throw_exception(status, "open");
  is_open = true;
}

ftc::~ftc()
{
  if(sock != -1)
    ftc_detail::close(sock);
#ifdef TEC_WINDOWS
  WSACleanup();
#endif
}

void ftc::close()
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");

  char command = 2;
  ftc_detail::send(sock, &command, sizeof(command));
  int status = ftc_detail::recv_status(sock);
  if(status < 0)
    ftc_detail::throw_exception(status, "close");
  is_open = false;
}

bool ftc::isOpen()
{
  return is_open;
}

void ftc::setPosition( unsigned long long p)
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");

  char buf_start[1+8] = {5};
  std::memcpy(&buf_start[1], static_cast<const char*>(static_cast<const void*>(&p))
              , (std::min<std::size_t>)(sizeof(p), 8u));
  ftc_detail::endian_swap(&buf_start[1]
                          , (std::min<std::size_t>)(8u, sizeof(p)));
  ftc_detail::send(sock, buf_start, sizeof(buf_start));
  int status = ftc_detail::recv_status(sock);
  if(status < 0)
    ftc_detail::throw_exception(status, "setPosition");
}

void ftc::setReadBufferSize( unsigned long long size )
{
}

unsigned long long ftc::getPosition()
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");
  return ftc_detail::return_long_long_command(sock, 4);
}

unsigned long long ftc::getReadBufferSize()
{
  return 0;
}

void ftc::setSize(unsigned long long s)
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");
  else if(!writable)
    throw NoPermissionException("mutable operation on read-only file");
  char buf_start[1+8] = {3};
  std::memcpy(&buf_start[1], static_cast<const char*>(static_cast<const void*>(&s))
              , (std::min<std::size_t>)(sizeof(s), 8u));
  ftc_detail::endian_swap(&buf_start[1]
                          , (std::min<std::size_t>)(8u, sizeof(s)));
  ftc_detail::send(sock, buf_start, sizeof(buf_start));
  int status = ftc_detail::recv_status(sock);
  if(status < 0)
    ftc_detail::throw_exception(status, "setSize");
}

unsigned long long ftc::getSize() const
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");
  return ftc_detail::return_long_long_command(sock, 6);
}

unsigned long long ftc::read(char* data, unsigned long long size, unsigned long long p)
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");
  char buf_start[1+8+8] = {7};
  std::memcpy(&buf_start[1], static_cast<const char*>(static_cast<const void*>(&p))
              , (std::min<std::size_t>)(sizeof(p), sizeof(buf_start)-1));
  std::memcpy(&buf_start[1+8], static_cast<const char*>(static_cast<const void*>(&size))
              , (std::min<std::size_t>)(sizeof(size), sizeof(buf_start)-1));
  ftc_detail::endian_swap(&buf_start[1]
                          , (std::min<std::size_t>)(8u, sizeof(p)));
  ftc_detail::endian_swap(&buf_start[1+8]
                          , (std::min<std::size_t>)(8u, sizeof(p)));
  ftc_detail::send(sock, buf_start, sizeof(buf_start));

  char* current = data;
  char* last = data + size;
  while(current != last)
  {
#ifdef TEC_WINDOWS
    WSABUF buffers = {last-current, current};
    DWORD count = 0, flags = 0;
    bool retry = false;
    int r = 0;
    do
    {
      r = ::WSARecv(sock, &buffers, 1, &count, &flags, 0, 0);
      if(r == -1 && WSAGetLastError() == WSAENOBUFS
         && buffers.len > ftc_detail::win32_minimum_buffer_size*2)
      {
        buffers.len /= 2;
        retry = true;
      }
      else
         retry = false;
    } while(retry);
    if(r != 0)
    {
      ftc_detail::errno_buffer b = ftc_detail::errno_message();
      ftc_detail::throw_exception(b.message());
    }
    else if(count == 0)
      return current - data;
    current += count;
#else
    ssize_t r = 0;
    do
    {
      r = ::recv(sock, current, last-current, 0);
    } while(r == -1 && errno == EINTR);
    if(r == -1)
    {
      ftc_detail::errno_buffer b = ftc_detail::errno_message();
      ftc_detail::throw_exception(b.message());
    }
    else if(r == 0)
      return current - data;
    current += r;
#endif
    assert(current <= last);
  }

  return current-data;
}

unsigned long long ftc::write(char const* data, unsigned long long size, unsigned long long p)
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");
  else if(!writable)
    throw NoPermissionException("mutable operation on read-only file");
  char buf_start[1+8+8] = {8};
  std::memcpy(&buf_start[1], static_cast<const char*>(static_cast<const void*>(&p))
              , (std::min<std::size_t>)(sizeof(p), 8u));
  std::memcpy(&buf_start[1+8], static_cast<const char*>(static_cast<const void*>(&size))
              , (std::min<std::size_t>)(sizeof(size), 8u));
  ftc_detail::endian_swap(&buf_start[1]
                          , (std::min<std::size_t>)(8u, sizeof(p)));
  ftc_detail::endian_swap(&buf_start[1+8]
                          , (std::min<std::size_t>)(8u, sizeof(size)));
  ftc_detail::send(sock, buf_start, sizeof(buf_start));

  const char* current = data
    , *last = data + size;
  while(current != last)
  {
#ifdef TEC_WINDOWS
    WSABUF buffers = {last-current, const_cast<char*>(current)};
    DWORD count = 0;
    bool retry = false;
    int r = 0;
    do
    {
      r = ::WSASend(sock, &buffers, 1, &count, 0, 0, 0);
      if(r == -1 && WSAGetLastError() == WSAENOBUFS
         && buffers.len > ftc_detail::win32_minimum_buffer_size*2)
      {
        buffers.len /= 2;
        retry = true;
      }
      else
         retry = false;
    } while(retry);
    if(r != 0)
    {
      ftc_detail::errno_buffer b = ftc_detail::errno_message();
      ftc_detail::throw_exception(b.message());
    }
    else if(count == 0)
      return current - data;
    current += count;
#else
    ssize_t r = 0;
    do
    {
      r = ::send(sock, current, last-current, 0);
    } while(r == -1 && errno == EINTR);
    if(r < 0)
    {
      ftc_detail::errno_buffer b = ftc_detail::errno_message();
      ftc_detail::throw_exception(b.message());
    }
    else if(r == 0)
      return current - data;
    current += r;
#endif
    assert(current <= last);
  }

  int status = ftc_detail::recv_status(sock);
  if(status < 0)
    ftc_detail::throw_exception(status, "write");

  return current-data;
}

unsigned long long ftc::transferTo(unsigned long long position, unsigned long long nbytes, FILE* fd, char* buffer)
{
  if(!is_open)
    throw FileNotOpenException("file not open - close");

  if(nbytes == 0)
    return 0;

  std::vector<char> new_buffer((std::min)(1024u*1024u, static_cast<unsigned int>(nbytes)));

  unsigned long long bytes_to_read = nbytes;

  while(bytes_to_read != 0)
  {
    unsigned long long r = read(&new_buffer[0], new_buffer.size(), position);
    if(r == 0)
      return nbytes - bytes_to_read;
    position += r;
    bytes_to_read -= r;
#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4244) // warning C4244: 'argument' : 
                              // conversion from 'unsigned __int64' to 'size_t', possible loss of data
#endif
    fwrite(&new_buffer[0], r, 1, fd);
#ifdef _MSC_VER
#pragma warning(pop)
#endif
  }

  return nbytes;
}


