/*
* tests/TestSuite.h
*/

#ifndef TESTSUITE_H_
#define TESTSUITE_H_

#include "config.h"
#include <cxxtest/TestSuite.h>
#include <time.h>
#include <typeinfo>
#include <iostream>
#include <iterator>
#include <string>
#include <vector>
#if !defined(_WIN32) && !defined(WIN32) && !defined(_WINNT_WINVER)
#include <sys/time.h>
#endif
#include <openssl/md5.h>

#include "ftc.h"

#define KEY_SIZE 16
#define LARGESIZE 4*1024*1024*1024ull
#define TEC_TEST_FILENAME "/test";
#define TEC_TEST_NOT_FOUND_FILENAME "/FILE_NOT_FOUND";
#if !defined(_WIN32) && !defined(WIN32) && !defined(_WINNT_WINVER)
#define TEC_LOCAL_TEST_FILENAME "/test";
#else
#define TEC_LOCAL_TEST_FILENAME "\\test";
#endif

using namespace std;

class TestSuite: public CxxTest::TestSuite {
  private:
    unsigned char md5_key[KEY_SIZE];

    double diffTime( struct timeval* t_start, struct timeval* t_finish )
    {
      return t_finish->tv_sec - t_start->tv_sec + (t_finish->tv_usec - t_start->tv_usec) / 1.e6 ;
    }

  public:
    void setUP() { }

    void testConstructor()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
	  
	  MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
        ftc ch(filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
    }

    void testOpenFILE_NOT_FOUND()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_NOT_FOUND_FILENAME;
	  
      MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
      
      ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS(ch.open( true ), InvalidKeyException) ;
      TS_ASSERT( !ch.isOpen() ) ;
      TS_ASSERT_THROWS( ch.close(), FileNotOpenException ) ;
    }

    void testOpentest()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
      std::cout << "trying " << filename << std::endl;
	  MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
      ftc ch( filename.c_str(), filename.size(), false, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS_NOTHING( ch.open( true ) ) ;
      TS_ASSERT( ch.isOpen() ) ;
      TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
    }

    void testWritableANDReadOnlyFALSE()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
	  MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
      ftc ch( filename.c_str(), filename.size(), false, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS( ch.open( false ), NoPermissionException ) ;
      TS_ASSERT( !ch.isOpen() ) ;
    }

    void testCLOSE()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
	  MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
      ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS( ch.close(), FileNotOpenException ) ;
      TS_ASSERT_THROWS_NOTHING( ch.open( false ) ) ;
      TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
    }

    void testSIZE()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
	  MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
      ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS_NOTHING( ch.open( false ) ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setSize( 0 ) ) ;
      TS_ASSERT_EQUALS( ch.getSize(),  0ull ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setSize( 10 ) ) ;
      TS_ASSERT_EQUALS( ch.getSize(),  10ull ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setSize( 1024*1024 ) ) ;
      TS_ASSERT_EQUALS( ch.getSize(),  1024*1024ull ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setSize( 0 ) ) ;
      TS_ASSERT_EQUALS( ch.getSize(),  0ull ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setSize( LARGESIZE ) ) ;
      TS_ASSERT_EQUALS( ch.getSize(),  LARGESIZE ) ;
      TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
    }

    void testSetSizeReadOnly()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
	  MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
      ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS_NOTHING( ch.open( true ) ) ;
      TS_ASSERT_THROWS( ch.setSize( 10 ), NoPermissionException ) ;
      TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
    }

    void testSETPOSITION()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME ;
	  MD5((const unsigned char*)filename.c_str(),filename.size(),md5_key);
      ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS_NOTHING( ch.open( false ) ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setPosition( 1024 ) ) ;
      TS_ASSERT_EQUALS( ch.getPosition(), 1024ull ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setPosition( 0 ) ) ;
      TS_ASSERT_EQUALS( ch.getPosition(), 0ull ) ;
      TS_ASSERT_THROWS_NOTHING( ch.setPosition( LARGESIZE ) ) ;
      TS_ASSERT_EQUALS( ch.getPosition(), LARGESIZE ) ;
      TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
    }

   void testReadAndWrite()
   {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
      unsigned long long nbytes = filename.size();
	  MD5((const unsigned char*)filename.c_str(), nbytes ,md5_key);
      ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS_NOTHING( ch.open( false ) ) ;
      
      TS_ASSERT_EQUALS(ch.write( filename.c_str(), nbytes , 0 ) , nbytes )  ;
      
      std::vector<char> data(nbytes);
      TS_ASSERT_EQUALS(ch.read( &data[0], nbytes, 0 ), nbytes) ;
      //for(unsigned int i = 0 ; i < nbytes ; i++)
      //    cout << data[i] << endl;
      std::string temp(data.begin(), data.end());
      TS_ASSERT_EQUALS(filename, temp) ;
      TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
   }

    void testReadAndWriteAfter3GB()
    {
      string filename(SERVER_TMP_PATH);  
      filename += TEC_TEST_FILENAME;
      unsigned long long nbytes = filename.size();
	  MD5((const unsigned char*)filename.c_str(), nbytes ,md5_key);
      ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
      TS_ASSERT_THROWS_NOTHING( ch.open( false ) ) ;
      
      TS_ASSERT_EQUALS(ch.write( filename.c_str(), nbytes , LARGESIZE ) , nbytes )  ;
      
      std::vector<char> data(nbytes);
      TS_ASSERT_EQUALS(ch.read( &data[0], nbytes, LARGESIZE ), nbytes) ;
      std::string temp(data.begin(), data.end());
      TS_ASSERT_EQUALS(filename, temp) ;
      TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
    }

    void testTransferTo()
    {
      std::string filename = SERVER_TMP_PATH;
      filename += TEC_TEST_FILENAME;

      unsigned long long size = filename.size();
      MD5(static_cast<unsigned char const*>(static_cast<void const*>(filename.c_str())), size, md5_key);
      ftc ch(filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, static_cast<char const*>(static_cast<void const*>(md5_key)), KEY_SIZE);
      
      TS_ASSERT_THROWS_NOTHING(ch.open(false));

      TS_ASSERT_THROWS_NOTHING(ch.setSize(0));
      TS_ASSERT_EQUALS(ch.write(filename.c_str(), size, 0), size);

      std::string temp_file = LOCAL_TMP_PATH;
      temp_file += "/temp_file";
      FILE* f = fopen(temp_file.c_str(), "w+");
      TS_ASSERT(f != 0);

      TS_ASSERT_EQUALS(ch.transferTo(0, size, f, 0), size);
      
      std::vector<char> readed(size);

      fseek(f, 0, SEEK_SET);
      fread(&readed[0], size, 1, f);

      TS_ASSERT_EQUALS(std::equal(readed.begin(), readed.end(), filename.c_str()), true);

      fclose(f);
    }

   void testOverwrite()
   {
     string filename(SERVER_TMP_PATH);  
     filename += TEC_TEST_FILENAME;
     unsigned long long nbytes = filename.size();
     MD5((const unsigned char*)filename.c_str(), nbytes ,md5_key);
     ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
     TS_ASSERT_THROWS_NOTHING( ch.open( false ) ) ;
      
     TS_ASSERT_EQUALS(ch.write( filename.c_str(), nbytes , 0 ) , nbytes )  ;
      
     std::vector<char> data(nbytes);
     TS_ASSERT_EQUALS(ch.read( &data[0], nbytes, 0 ), nbytes) ;
     std::string temp(data.begin(), data.end());
     TS_ASSERT_EQUALS(filename, temp) ;

     TS_ASSERT_EQUALS(ch.write( filename.c_str(), nbytes , 1 ) , nbytes )  ;
     {
       std::vector<char> data(nbytes+1);
       TS_ASSERT_EQUALS(ch.read( &data[0], nbytes+1, 0 ), nbytes+1) ;
       std::string temp(data.begin(), data.end());
       std::string filename_temp;
       filename_temp += filename[0];
       filename_temp += filename;
       TS_ASSERT_EQUALS(filename_temp, temp) ;
     }

     TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
   }

   void testSizeAfterOverwrite()
   {
     string filename(SERVER_TMP_PATH);  
     filename += TEC_TEST_FILENAME;
     unsigned long long nbytes = filename.size();
     MD5((const unsigned char*)filename.c_str(), nbytes ,md5_key);
     ftc ch( filename.c_str(), filename.size(), true, SERVER_HOST, SERVER_PORT, (const char*)md5_key, KEY_SIZE) ;
     TS_ASSERT_THROWS_NOTHING( ch.open( false ) ) ;
      
     TS_ASSERT_EQUALS(ch.write( filename.c_str(), nbytes , 0 ) , nbytes )  ;
     TS_ASSERT_EQUALS(ch.write( filename.c_str(), nbytes , 1 ) , nbytes )  ;

     TS_ASSERT_EQUALS(ch.getSize(), nbytes+1);

     TS_ASSERT_THROWS_NOTHING( ch.close() ) ;
   }

} ;

#endif
